"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments)).next());
    });
};
const constants_1 = require("../common/constants");
class RunCommandBase {
    constructor($platformService, $liveSyncService, $projectData, $options, $emulatorPlatformService, $devicePlatformsConstants, $errors, $devicesService, $hostInfo, $iosDeviceOperations, $mobileHelper, $platformsData) {
        this.$platformService = $platformService;
        this.$liveSyncService = $liveSyncService;
        this.$projectData = $projectData;
        this.$options = $options;
        this.$emulatorPlatformService = $emulatorPlatformService;
        this.$devicePlatformsConstants = $devicePlatformsConstants;
        this.$errors = $errors;
        this.$devicesService = $devicesService;
        this.$hostInfo = $hostInfo;
        this.$iosDeviceOperations = $iosDeviceOperations;
        this.$mobileHelper = $mobileHelper;
        this.$platformsData = $platformsData;
        this.allowedParameters = [];
    }
    execute(args) {
        return __awaiter(this, void 0, void 0, function* () {
            return this.executeCore(args);
        });
    }
    canExecute(args) {
        return __awaiter(this, void 0, void 0, function* () {
            if (args.length) {
                this.$errors.fail(constants_1.ERROR_NO_VALID_SUBCOMMAND_FORMAT, "run");
            }
            this.$projectData.initializeProjectData();
            if (!this.platform && !this.$hostInfo.isDarwin) {
                this.platform = this.$devicePlatformsConstants.Android;
            }
            const availablePlatforms = this.platform ? [this.platform] : this.$platformsData.availablePlatforms;
            for (let platform of availablePlatforms) {
                const platformData = this.$platformsData.getPlatformData(platform, this.$projectData);
                const platformProjectService = platformData.platformProjectService;
                yield platformProjectService.validate(this.$projectData);
            }
            return true;
        });
    }
    executeCore(args) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.$options.bundle) {
                this.$options.watch = false;
            }
            yield this.$devicesService.initialize({
                deviceId: this.$options.device,
                platform: this.platform,
                emulator: this.$options.emulator,
                skipDeviceDetectionInterval: true,
                skipInferPlatform: !this.platform
            });
            yield this.$devicesService.detectCurrentlyAttachedDevices();
            const devices = this.$devicesService.getDeviceInstances();
            const deviceDescriptors = devices.filter(d => !this.platform || d.deviceInfo.platform === this.platform)
                .map(d => {
                const info = {
                    identifier: d.deviceInfo.identifier,
                    buildAction: () => __awaiter(this, void 0, void 0, function* () {
                        const buildConfig = {
                            buildForDevice: !d.isEmulator,
                            projectDir: this.$options.path,
                            clean: this.$options.clean,
                            teamId: this.$options.teamId,
                            device: this.$options.device,
                            provision: this.$options.provision,
                            release: this.$options.release,
                            keyStoreAlias: this.$options.keyStoreAlias,
                            keyStorePath: this.$options.keyStorePath,
                            keyStoreAliasPassword: this.$options.keyStoreAliasPassword,
                            keyStorePassword: this.$options.keyStorePassword
                        };
                        yield this.$platformService.buildPlatform(d.deviceInfo.platform, buildConfig, this.$projectData);
                        const pathToBuildResult = yield this.$platformService.lastOutputPath(d.deviceInfo.platform, buildConfig, this.$projectData);
                        return pathToBuildResult;
                    })
                };
                return info;
            });
            const workingWithiOSDevices = !this.platform || this.$mobileHelper.isiOSPlatform(this.platform);
            const shouldKeepProcessAlive = this.$options.watch || !this.$options.justlaunch;
            if (workingWithiOSDevices && shouldKeepProcessAlive) {
                this.$iosDeviceOperations.setShouldDispose(false);
            }
            if (this.$options.release || this.$options.bundle) {
                const runPlatformOptions = {
                    device: this.$options.device,
                    emulator: this.$options.emulator,
                    justlaunch: this.$options.justlaunch
                };
                const deployOptions = _.merge({
                    projectDir: this.$projectData.projectDir,
                    clean: true,
                }, this.$options.argv);
                yield this.$platformService.deployPlatform(args[0], this.$options, deployOptions, this.$projectData, this.$options);
                yield this.$platformService.startApplication(args[0], runPlatformOptions, this.$projectData.projectId);
                return this.$platformService.trackProjectType(this.$projectData);
            }
            const liveSyncInfo = {
                projectDir: this.$projectData.projectDir,
                skipWatcher: !this.$options.watch,
                watchAllFiles: this.$options.syncAllFiles,
                clean: this.$options.clean
            };
            yield this.$liveSyncService.liveSync(deviceDescriptors, liveSyncInfo);
        });
    }
}
exports.RunCommandBase = RunCommandBase;
$injector.registerCommand("run|*all", RunCommandBase);
class RunIosCommand extends RunCommandBase {
    constructor($platformService, $platformsData, $devicePlatformsConstants, $errors, $liveSyncService, $projectData, $options, $emulatorPlatformService, $devicesService, $hostInfo, $iosDeviceOperations, $mobileHelper) {
        super($platformService, $liveSyncService, $projectData, $options, $emulatorPlatformService, $devicePlatformsConstants, $errors, $devicesService, $hostInfo, $iosDeviceOperations, $mobileHelper, $platformsData);
        this.$platformsData = $platformsData;
        this.$devicePlatformsConstants = $devicePlatformsConstants;
        this.$errors = $errors;
        this.allowedParameters = [];
    }
    get platform() {
        return this.$devicePlatformsConstants.iOS;
    }
    execute(args) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.$platformService.isPlatformSupportedForOS(this.$devicePlatformsConstants.iOS, this.$projectData)) {
                this.$errors.fail(`Applications for platform ${this.$devicePlatformsConstants.iOS} can not be built on this OS`);
            }
            return this.executeCore([this.$platformsData.availablePlatforms.iOS]);
        });
    }
    canExecute(args) {
        const _super = name => super[name];
        return __awaiter(this, void 0, void 0, function* () {
            return (yield _super("canExecute").call(this, args)) && (yield this.$platformService.validateOptions(this.$options.provision, this.$projectData, this.$platformsData.availablePlatforms.iOS));
        });
    }
}
exports.RunIosCommand = RunIosCommand;
$injector.registerCommand("run|ios", RunIosCommand);
class RunAndroidCommand extends RunCommandBase {
    constructor($platformService, $platformsData, $devicePlatformsConstants, $errors, $liveSyncService, $projectData, $options, $emulatorPlatformService, $devicesService, $hostInfo, $iosDeviceOperations, $mobileHelper) {
        super($platformService, $liveSyncService, $projectData, $options, $emulatorPlatformService, $devicePlatformsConstants, $errors, $devicesService, $hostInfo, $iosDeviceOperations, $mobileHelper, $platformsData);
        this.$platformsData = $platformsData;
        this.$devicePlatformsConstants = $devicePlatformsConstants;
        this.$errors = $errors;
        this.allowedParameters = [];
    }
    get platform() {
        return this.$devicePlatformsConstants.Android;
    }
    execute(args) {
        return __awaiter(this, void 0, void 0, function* () {
            return this.executeCore([this.$platformsData.availablePlatforms.Android]);
        });
    }
    canExecute(args) {
        const _super = name => super[name];
        return __awaiter(this, void 0, void 0, function* () {
            yield _super("canExecute").call(this, args);
            if (!this.$platformService.isPlatformSupportedForOS(this.$devicePlatformsConstants.Android, this.$projectData)) {
                this.$errors.fail(`Applications for platform ${this.$devicePlatformsConstants.Android} can not be built on this OS`);
            }
            if (this.$options.release && (!this.$options.keyStorePath || !this.$options.keyStorePassword || !this.$options.keyStoreAlias || !this.$options.keyStoreAliasPassword)) {
                this.$errors.fail("When producing a release build, you need to specify all --key-store-* options.");
            }
            return this.$platformService.validateOptions(this.$options.provision, this.$projectData, this.$platformsData.availablePlatforms.Android);
        });
    }
}
exports.RunAndroidCommand = RunAndroidCommand;
$injector.registerCommand("run|android", RunAndroidCommand);
