"use strict";
var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments)).next());
    });
};
const path = require("path");
const choki = require("chokidar");
const events_1 = require("events");
const helpers_1 = require("../../common/helpers");
const constants_1 = require("../../constants");
const constants_2 = require("../../common/constants");
const deviceDescriptorPrimaryKey = "identifier";
const LiveSyncEvents = {
    liveSyncStopped: "liveSyncStopped",
    liveSyncError: "liveSyncError",
    liveSyncExecuted: "liveSyncExecuted",
    liveSyncStarted: "liveSyncStarted",
    liveSyncNotification: "notify"
};
class LiveSyncService extends events_1.EventEmitter {
    constructor($platformService, $projectDataService, $devicesService, $mobileHelper, $devicePlatformsConstants, $nodeModulesDependenciesBuilder, $logger, $processService, $hooksService, $pluginsService, $injector) {
        super();
        this.$platformService = $platformService;
        this.$projectDataService = $projectDataService;
        this.$devicesService = $devicesService;
        this.$mobileHelper = $mobileHelper;
        this.$devicePlatformsConstants = $devicePlatformsConstants;
        this.$nodeModulesDependenciesBuilder = $nodeModulesDependenciesBuilder;
        this.$logger = $logger;
        this.$processService = $processService;
        this.$hooksService = $hooksService;
        this.$pluginsService = $pluginsService;
        this.$injector = $injector;
        this.liveSyncProcessesInfo = {};
    }
    liveSync(deviceDescriptors, liveSyncData) {
        return __awaiter(this, void 0, void 0, function* () {
            const projectData = this.$projectDataService.getProjectData(liveSyncData.projectDir);
            yield this.$pluginsService.ensureAllDependenciesAreInstalled(projectData);
            yield this.liveSyncOperation(deviceDescriptors, liveSyncData, projectData);
        });
    }
    stopLiveSync(projectDir, deviceIdentifiers) {
        return __awaiter(this, void 0, void 0, function* () {
            const liveSyncProcessInfo = this.liveSyncProcessesInfo[projectDir];
            if (liveSyncProcessInfo) {
                _.each(deviceIdentifiers, deviceId => {
                    _.remove(liveSyncProcessInfo.deviceDescriptors, descriptor => {
                        const shouldRemove = descriptor.identifier === deviceId;
                        if (shouldRemove) {
                            this.emit(LiveSyncEvents.liveSyncStopped, { projectDir, deviceIdentifier: descriptor.identifier });
                        }
                        return shouldRemove;
                    });
                });
                if (!deviceIdentifiers || !deviceIdentifiers.length || !liveSyncProcessInfo.deviceDescriptors || !liveSyncProcessInfo.deviceDescriptors.length) {
                    if (liveSyncProcessInfo.timer) {
                        clearTimeout(liveSyncProcessInfo.timer);
                    }
                    if (liveSyncProcessInfo.watcherInfo && liveSyncProcessInfo.watcherInfo.watcher) {
                        liveSyncProcessInfo.watcherInfo.watcher.close();
                    }
                    liveSyncProcessInfo.watcherInfo = null;
                    if (liveSyncProcessInfo.actionsChain) {
                        yield liveSyncProcessInfo.actionsChain;
                    }
                    _.each(liveSyncProcessInfo.deviceDescriptors, descriptor => {
                        this.emit(LiveSyncEvents.liveSyncStopped, { projectDir, deviceIdentifier: descriptor.identifier });
                    });
                    liveSyncProcessInfo.isStopped = true;
                    liveSyncProcessInfo.deviceDescriptors = [];
                    const projectData = this.$projectDataService.getProjectData(projectDir);
                    yield this.$hooksService.executeAfterHooks('watch', {
                        hookArgs: {
                            projectData
                        }
                    });
                }
            }
        });
    }
    refreshApplication(projectData, liveSyncResultInfo) {
        return __awaiter(this, void 0, void 0, function* () {
            const platformLiveSyncService = this.getLiveSyncService(liveSyncResultInfo.deviceAppData.platform);
            try {
                yield platformLiveSyncService.refreshApplication(projectData, liveSyncResultInfo);
            }
            catch (err) {
                this.$logger.info(`Error while trying to start application ${projectData.projectId} on device ${liveSyncResultInfo.deviceAppData.device.deviceInfo.identifier}. Error is: ${err.message || err}`);
                const msg = `Unable to start application ${projectData.projectId} on device ${liveSyncResultInfo.deviceAppData.device.deviceInfo.identifier}. Try starting it manually.`;
                this.$logger.warn(msg);
                this.emit(LiveSyncEvents.liveSyncNotification, {
                    projectDir: projectData.projectDir,
                    applicationIdentifier: projectData.projectId,
                    deviceIdentifier: liveSyncResultInfo.deviceAppData.device.deviceInfo.identifier,
                    notification: msg
                });
            }
            this.emit(LiveSyncEvents.liveSyncExecuted, {
                projectDir: projectData.projectDir,
                applicationIdentifier: projectData.projectId,
                syncedFiles: liveSyncResultInfo.modifiedFilesData.map(m => m.getLocalPath()),
                deviceIdentifier: liveSyncResultInfo.deviceAppData.device.deviceInfo.identifier
            });
            this.$logger.info(`Successfully synced application ${liveSyncResultInfo.deviceAppData.appIdentifier} on device ${liveSyncResultInfo.deviceAppData.device.deviceInfo.identifier}.`);
        });
    }
    liveSyncOperation(deviceDescriptors, liveSyncData, projectData) {
        return __awaiter(this, void 0, void 0, function* () {
            const isAlreadyLiveSyncing = this.liveSyncProcessesInfo[projectData.projectDir] && !this.liveSyncProcessesInfo[projectData.projectDir].isStopped;
            this.setLiveSyncProcessInfo(liveSyncData.projectDir, deviceDescriptors);
            const deviceDescriptorsForInitialSync = isAlreadyLiveSyncing ? _.differenceBy(deviceDescriptors, this.liveSyncProcessesInfo[projectData.projectDir].deviceDescriptors, deviceDescriptorPrimaryKey) : deviceDescriptors;
            yield this.initialSync(projectData, deviceDescriptorsForInitialSync, liveSyncData);
            if (!liveSyncData.skipWatcher && deviceDescriptors && deviceDescriptors.length) {
                this.$injector.resolve("usbLiveSyncService").isInitialized = true;
                yield this.startWatcher(projectData, liveSyncData);
            }
        });
    }
    setLiveSyncProcessInfo(projectDir, deviceDescriptors) {
        this.liveSyncProcessesInfo[projectDir] = this.liveSyncProcessesInfo[projectDir] || Object.create(null);
        this.liveSyncProcessesInfo[projectDir].actionsChain = this.liveSyncProcessesInfo[projectDir].actionsChain || Promise.resolve();
        this.liveSyncProcessesInfo[projectDir].isStopped = false;
        const currentDeviceDescriptors = this.liveSyncProcessesInfo[projectDir].deviceDescriptors || [];
        this.liveSyncProcessesInfo[projectDir].deviceDescriptors = _.uniqBy(currentDeviceDescriptors.concat(deviceDescriptors), deviceDescriptorPrimaryKey);
    }
    getLiveSyncService(platform) {
        if (this.$mobileHelper.isiOSPlatform(platform)) {
            return this.$injector.resolve("iOSLiveSyncService");
        }
        else if (this.$mobileHelper.isAndroidPlatform(platform)) {
            return this.$injector.resolve("androidLiveSyncService");
        }
        throw new Error(`Invalid platform ${platform}. Supported platforms are: ${this.$mobileHelper.platformNames.join(", ")}`);
    }
    ensureLatestAppPackageIsInstalledOnDevice(options) {
        return __awaiter(this, void 0, void 0, function* () {
            const platform = options.device.deviceInfo.platform;
            if (options.preparedPlatforms.indexOf(platform) === -1) {
                options.preparedPlatforms.push(platform);
                yield this.$platformService.preparePlatform(platform, {
                    bundle: false,
                    release: false,
                }, null, options.projectData, {}, options.modifiedFiles);
            }
            const rebuildInfo = _.find(options.rebuiltInformation, info => info.isEmulator === options.device.isEmulator && info.platform === platform);
            if (rebuildInfo) {
                yield this.$platformService.installApplication(options.device, { release: false }, options.projectData, rebuildInfo.pathToBuildItem, options.deviceBuildInfoDescriptor.outputPath);
                return;
            }
            const shouldBuild = yield this.$platformService.shouldBuild(platform, options.projectData, { buildForDevice: !options.device.isEmulator, clean: options.liveSyncData && options.liveSyncData.clean }, options.deviceBuildInfoDescriptor.outputPath);
            let pathToBuildItem = null;
            let action = constants_1.LiveSyncTrackActionNames.LIVESYNC_OPERATION;
            if (shouldBuild) {
                pathToBuildItem = yield options.deviceBuildInfoDescriptor.buildAction();
                options.rebuiltInformation.push({ isEmulator: options.device.isEmulator, platform, pathToBuildItem });
                action = constants_1.LiveSyncTrackActionNames.LIVESYNC_OPERATION_BUILD;
            }
            if (!options.settings[platform][options.device.deviceInfo.type]) {
                let isForDevice = !options.device.isEmulator;
                options.settings[platform][options.device.deviceInfo.type] = true;
                if (this.$mobileHelper.isAndroidPlatform(platform)) {
                    options.settings[platform][constants_2.DeviceTypes.Emulator] = true;
                    options.settings[platform][constants_2.DeviceTypes.Device] = true;
                    isForDevice = null;
                }
                yield this.$platformService.trackActionForPlatform({ action, platform, isForDevice });
            }
            yield this.$platformService.trackActionForPlatform({ action: constants_1.LiveSyncTrackActionNames.DEVICE_INFO, platform, isForDevice: !options.device.isEmulator, deviceOsVersion: options.device.deviceInfo.version });
            const shouldInstall = yield this.$platformService.shouldInstall(options.device, options.projectData, options.deviceBuildInfoDescriptor.outputPath);
            if (shouldInstall) {
                yield this.$platformService.installApplication(options.device, { release: false }, options.projectData, pathToBuildItem, options.deviceBuildInfoDescriptor.outputPath);
            }
        });
    }
    initialSync(projectData, deviceDescriptors, liveSyncData) {
        return __awaiter(this, void 0, void 0, function* () {
            const preparedPlatforms = [];
            const rebuiltInformation = [];
            const settings = this.getDefaultLatestAppPackageInstalledSettings();
            const deviceAction = (device) => __awaiter(this, void 0, void 0, function* () {
                try {
                    this.emit(LiveSyncEvents.liveSyncStarted, {
                        projectDir: projectData.projectDir,
                        deviceIdentifier: device.deviceInfo.identifier,
                        applicationIdentifier: projectData.projectId
                    });
                    const platform = device.deviceInfo.platform;
                    const deviceBuildInfoDescriptor = _.find(deviceDescriptors, dd => dd.identifier === device.deviceInfo.identifier);
                    yield this.ensureLatestAppPackageIsInstalledOnDevice({
                        device,
                        preparedPlatforms,
                        rebuiltInformation,
                        projectData,
                        deviceBuildInfoDescriptor,
                        liveSyncData,
                        settings
                    });
                    const liveSyncResultInfo = yield this.getLiveSyncService(platform).fullSync({
                        projectData, device,
                        syncAllFiles: liveSyncData.watchAllFiles,
                        useLiveEdit: liveSyncData.useLiveEdit,
                        watch: !liveSyncData.skipWatcher
                    });
                    yield this.$platformService.trackActionForPlatform({ action: "LiveSync", platform: device.deviceInfo.platform, isForDevice: !device.isEmulator, deviceOsVersion: device.deviceInfo.version });
                    yield this.refreshApplication(projectData, liveSyncResultInfo);
                }
                catch (err) {
                    this.$logger.warn(`Unable to apply changes on device: ${device.deviceInfo.identifier}. Error is: ${err.message}.`);
                    this.emit(LiveSyncEvents.liveSyncError, {
                        error: err,
                        deviceIdentifier: device.deviceInfo.identifier,
                        projectDir: projectData.projectDir,
                        applicationIdentifier: projectData.projectId
                    });
                    yield this.stopLiveSync(projectData.projectDir, [device.deviceInfo.identifier]);
                }
            });
            yield this.addActionToChain(projectData.projectDir, () => this.$devicesService.execute(deviceAction, (device) => _.some(deviceDescriptors, deviceDescriptor => deviceDescriptor.identifier === device.deviceInfo.identifier)));
        });
    }
    getDefaultLatestAppPackageInstalledSettings() {
        return {
            [this.$devicePlatformsConstants.Android]: {
                [constants_2.DeviceTypes.Device]: false,
                [constants_2.DeviceTypes.Emulator]: false
            },
            [this.$devicePlatformsConstants.iOS]: {
                [constants_2.DeviceTypes.Device]: false,
                [constants_2.DeviceTypes.Emulator]: false
            }
        };
    }
    startWatcher(projectData, liveSyncData) {
        return __awaiter(this, void 0, void 0, function* () {
            let pattern = [constants_1.APP_FOLDER_NAME];
            if (liveSyncData.watchAllFiles) {
                const productionDependencies = this.$nodeModulesDependenciesBuilder.getProductionDependencies(projectData.projectDir);
                pattern.push(constants_1.PACKAGE_JSON_FILE_NAME);
                for (let index in productionDependencies) {
                    pattern.push(productionDependencies[index].directory);
                }
            }
            const currentWatcherInfo = this.liveSyncProcessesInfo[liveSyncData.projectDir].watcherInfo;
            if (!currentWatcherInfo || currentWatcherInfo.pattern !== pattern) {
                if (currentWatcherInfo) {
                    currentWatcherInfo.watcher.close();
                }
                let filesToSync = [], filesToRemove = [];
                let timeoutTimer;
                const startTimeout = () => {
                    timeoutTimer = setTimeout(() => __awaiter(this, void 0, void 0, function* () {
                        yield this.addActionToChain(projectData.projectDir, () => __awaiter(this, void 0, void 0, function* () {
                            if (filesToSync.length || filesToRemove.length) {
                                try {
                                    let currentFilesToSync = _.cloneDeep(filesToSync);
                                    filesToSync = [];
                                    let currentFilesToRemove = _.cloneDeep(filesToRemove);
                                    filesToRemove = [];
                                    const allModifiedFiles = [].concat(currentFilesToSync).concat(currentFilesToRemove);
                                    const preparedPlatforms = [];
                                    const rebuiltInformation = [];
                                    const latestAppPackageInstalledSettings = this.getDefaultLatestAppPackageInstalledSettings();
                                    yield this.$devicesService.execute((device) => __awaiter(this, void 0, void 0, function* () {
                                        const liveSyncProcessInfo = this.liveSyncProcessesInfo[projectData.projectDir];
                                        const deviceBuildInfoDescriptor = _.find(liveSyncProcessInfo.deviceDescriptors, dd => dd.identifier === device.deviceInfo.identifier);
                                        yield this.ensureLatestAppPackageIsInstalledOnDevice({
                                            device,
                                            preparedPlatforms,
                                            rebuiltInformation,
                                            projectData,
                                            deviceBuildInfoDescriptor,
                                            settings: latestAppPackageInstalledSettings,
                                            modifiedFiles: allModifiedFiles
                                        });
                                        const service = this.getLiveSyncService(device.deviceInfo.platform);
                                        const settings = {
                                            projectData,
                                            filesToRemove: currentFilesToRemove,
                                            filesToSync: currentFilesToSync,
                                            isRebuilt: !!_.find(rebuiltInformation, info => info.isEmulator === device.isEmulator && info.platform === device.deviceInfo.platform),
                                            syncAllFiles: liveSyncData.watchAllFiles,
                                            useLiveEdit: liveSyncData.useLiveEdit
                                        };
                                        const liveSyncResultInfo = yield service.liveSyncWatchAction(device, settings);
                                        yield this.refreshApplication(projectData, liveSyncResultInfo);
                                    }), (device) => {
                                        const liveSyncProcessInfo = this.liveSyncProcessesInfo[projectData.projectDir];
                                        return liveSyncProcessInfo && _.some(liveSyncProcessInfo.deviceDescriptors, deviceDescriptor => deviceDescriptor.identifier === device.deviceInfo.identifier);
                                    });
                                }
                                catch (err) {
                                    const allErrors = err.allErrors;
                                    if (allErrors && _.isArray(allErrors)) {
                                        for (let deviceError of allErrors) {
                                            this.$logger.warn(`Unable to apply changes for device: ${deviceError.deviceIdentifier}. Error is: ${deviceError.message}.`);
                                            this.emit(LiveSyncEvents.liveSyncError, {
                                                error: deviceError,
                                                deviceIdentifier: deviceError.deviceIdentifier,
                                                projectDir: projectData.projectDir,
                                                applicationIdentifier: projectData.projectId
                                            });
                                            yield this.stopLiveSync(projectData.projectDir, [deviceError.deviceIdentifier]);
                                        }
                                    }
                                }
                            }
                        }));
                    }), 250);
                    this.liveSyncProcessesInfo[liveSyncData.projectDir].timer = timeoutTimer;
                };
                yield this.$hooksService.executeBeforeHooks('watch', {
                    hookArgs: {
                        projectData
                    }
                });
                const watcherOptions = {
                    ignoreInitial: true,
                    cwd: liveSyncData.projectDir,
                    awaitWriteFinish: {
                        pollInterval: 100,
                        stabilityThreshold: 500
                    },
                    ignored: ["**/.*", ".*"]
                };
                const watcher = choki.watch(pattern, watcherOptions)
                    .on("all", (event, filePath) => __awaiter(this, void 0, void 0, function* () {
                    clearTimeout(timeoutTimer);
                    filePath = path.join(liveSyncData.projectDir, filePath);
                    this.$logger.trace(`Chokidar raised event ${event} for ${filePath}.`);
                    if (event === "add" || event === "addDir" || event === "change") {
                        filesToSync.push(filePath);
                    }
                    else if (event === "unlink" || event === "unlinkDir") {
                        filesToRemove.push(filePath);
                    }
                    if (path.extname(filePath) !== constants_2.FileExtensions.TYPESCRIPT_FILE) {
                        startTimeout();
                    }
                }));
                this.liveSyncProcessesInfo[liveSyncData.projectDir].watcherInfo = { watcher, pattern };
                this.liveSyncProcessesInfo[liveSyncData.projectDir].timer = timeoutTimer;
                this.$processService.attachToProcessExitSignals(this, () => {
                    _.keys(this.liveSyncProcessesInfo).forEach(projectDir => {
                        this.stopLiveSync(projectDir);
                    });
                });
                this.$devicesService.on("deviceLost", (device) => __awaiter(this, void 0, void 0, function* () {
                    yield this.stopLiveSync(projectData.projectDir, [device.deviceInfo.identifier]);
                }));
            }
        });
    }
    addActionToChain(projectDir, action) {
        return __awaiter(this, void 0, void 0, function* () {
            const liveSyncInfo = this.liveSyncProcessesInfo[projectDir];
            if (liveSyncInfo) {
                liveSyncInfo.actionsChain = liveSyncInfo.actionsChain.then(() => __awaiter(this, void 0, void 0, function* () {
                    if (!liveSyncInfo.isStopped) {
                        const res = yield action();
                        return res;
                    }
                }));
                const result = yield liveSyncInfo.actionsChain;
                return result;
            }
        });
    }
}
__decorate([
    helpers_1.hook("liveSync")
], LiveSyncService.prototype, "liveSyncOperation", null);
exports.LiveSyncService = LiveSyncService;
$injector.register("liveSyncService", LiveSyncService);
class DeprecatedUsbLiveSyncService {
    constructor() {
        this.isInitialized = false;
    }
}
exports.DeprecatedUsbLiveSyncService = DeprecatedUsbLiveSyncService;
$injector.register("usbLiveSyncService", DeprecatedUsbLiveSyncService);
